@@ -19,6 +19,8 @@
 import unittest
 
 from .test_re import *
+from .test_diff import *
+from .test_executable import *
 
 if __name__ == '__main__':
     unittest.main()
